#include <stdio.h>
#include <linux/fs.h>
#include <fcntl.h>

#define XNET_IOCTL_DEBUG 0x9901

int main(int argc,char **argv)
{
    int major;
    int  pppdev,debug;
    debug=0;
  
    if (argc==2) {
	sscanf(argv[1],"%d",&debug);
    }

    pppdev = open("/dev/xnet0",O_RDWR);
    if (pppdev < 0) {
	perror("Can't open /dev/xnet0");
	exit(1);
    }

    ioctl(pppdev,XNET_IOCTL_DEBUG,debug);
}
